@@ -18,7 +18,7 @@ from collections import defaultdict
 import attr
 
 from hypothesis.internal.compat import int_from_bytes, int_to_bytes
-from hypothesis.internal.conjecture.choicetree import ChoiceTree
+from hypothesis.internal.conjecture.choicetree import ChoiceTree, prefix_selection_order
 from hypothesis.internal.conjecture.data import ConjectureResult, Overrun, Status
 from hypothesis.internal.conjecture.floats import (
     DRAW_FLOAT_LABEL,
@@ -258,7 +258,7 @@ class Shrinker:
         accept.__name__ = fn.__name__
         return property(accept)
 
-    def __init__(self, engine, initial, predicate):
+    def __init__(self, engine, initial, predicate, allow_transition):
         """Create a shrinker for a particular engine, with a given starting
         point and predicate. When shrink() is called it will attempt to find an
         example for which predicate is True and which is strictly smaller than
@@ -267,8 +267,10 @@ class Shrinker:
         Note that initial is a ConjectureData object, and predicate
         takes ConjectureData objects.
         """
+        assert predicate is not None or allow_transition is not None
         self.engine = engine
-        self.__predicate = predicate
+        self.__predicate = predicate or (lambda data: True)
+        self.__allow_transition = allow_transition or (lambda source, destination: True)
         self.__derived_values = {}
         self.__pending_shrink_explanation = None
 
@@ -363,16 +365,15 @@ class Shrinker:
 
     def incorporate_test_data(self, data):
         """Takes a ConjectureData or Overrun object updates the current
-        shrink_target if this data represents an improvement over it,
-        returning True if it is."""
-        if data is Overrun or data is self.shrink_target:
+        shrink_target if this data represents an improvement over it."""
+        if data.status < Status.VALID or data is self.shrink_target:
             return
-        if self.__predicate(data) and sort_key(data.buffer) < sort_key(
-            self.shrink_target.buffer
+        if (
+            self.__predicate(data)
+            and sort_key(data.buffer) < sort_key(self.shrink_target.buffer)
+            and self.__allow_transition(self.shrink_target, data)
         ):
             self.update_shrink_target(data)
-            return True
-        return False
 
     def cached_test_function(self, buffer):
         """Returns a cached version of the underlying test function, so
@@ -570,9 +571,6 @@ class Shrinker:
             if 0 < len(successful_passes) < len(passes):
                 self.fixate_shrink_passes(successful_passes)
 
-        for sp in passes:
-            sp.fixed_point_at = self.shrink_target
-
     @property
     def buffer(self):
         return self.shrink_target.buffer
@@ -1043,6 +1041,7 @@ class Shrinker:
             if a < 0 or b > len(group):
                 return False
             regions = []
+
             for e in group[a:b]:
                 t = (e.start, e.end, replacement)
                 if not regions or t[0] >= regions[-1][1]:
@@ -1450,17 +1449,13 @@ class ShrinkPass:
     index = attr.ib()
     shrinker = attr.ib()
 
-    next_prefix = attr.ib(default=())
-    fixed_point_at = attr.ib(default=None)
+    last_prefix = attr.ib(default=())
     successes = attr.ib(default=0)
     calls = attr.ib(default=0)
     shrinks = attr.ib(default=0)
     deletions = attr.ib(default=0)
 
     def step(self):
-        if self.fixed_point_at is self.shrinker.shrink_target:
-            return False
-
         tree = self.shrinker.shrink_pass_choice_trees[self]
         if tree.exhausted:
             return False
@@ -1470,8 +1465,8 @@ class ShrinkPass:
         size = len(self.shrinker.shrink_target.buffer)
         self.shrinker.explain_next_call_as(self.name)
         try:
-            self.next_prefix = tree.step(
-                self.next_prefix,
+            self.last_prefix = tree.step(
+                prefix_selection_order(self.last_prefix),
                 lambda chooser: self.run_with_chooser(self.shrinker, chooser),
             )
         finally:
